#include <stdbool.h>
#include <stdarg.h>
#include <alloca.h>
#include <stdint.h>
#include "swdCommsPacket.h"
#include "cortex.h"
#include "memap.h"
#include "util.h"
#include "plat.h"
#include "swd.h"



#define SWD_APP_VER				0x01010000



void main(void)
{
	packet_sz_t maxPkt, maxWordsTxRx;
	uint8_t blFlags, blVer = 0;
	
	//get caps
	getUsbCaps(&maxPkt, &blFlags);
	if (blFlags & BL_CAP_BL_VERSION_REPORTING)
		blVer = blGetBlVersion();
	
	platInit(blFlags & BL_CAP_FLAGS_CDC_SUPPORTED, blVer);
	maxWordsTxRx = (packet_sz_t)(maxPkt - sizeof(struct CommsPacket) - sizeof(struct SwdCommsMemPacket)) / sizeof(uint32_t);
	
	while (1) {
		packet_sz_t replyPayloadLen = PACKET_RX_FAIL, cmdPayloadLen;
		uint_fast8_t i, j = 0;
		struct CommsPacket *cp;
		uint8_t ahbApIdx;
		
		platWork();
		usbWork();
		
		if (packetCanSend() && (cmdPayloadLen = packetRx(&cp)) != PACKET_RX_FAIL) {
			
			switch (cp->cmd) {
			
				case SWD_COMMS_CMD_VER_INFO: {
					struct SwdCommsVerInfoRespPacketV5* rsp = (struct SwdCommsVerInfoRespPacketV5*)cp->payload;
					
					if (cmdPayloadLen)
						break;
					
					rsp->swdAppVer = SWD_APP_VER;
					rsp->flags = platGetFlags();
					rsp->maxXferBytes = maxPkt;
					rsp->hwType = HW_TYPE;
					rsp->hwVer = platGetHwVerForComms();
					if (platGetFlags() & (PWR_FLAG_PWR_CTRL_SETTABLE | PWR_FLAG_PWR_CTRL_ON_OFF)) {
						
						uint16_t millivoltsMin, millivoltsMax, milliampsMax;	//canot pass pointer ot rsp since it mightbe unaligned
						
						platGetSupplyAbilities(&millivoltsMin, &millivoltsMax, &milliampsMax);
						rsp->millivoltsMin = millivoltsMin;
						rsp->millivoltsMax = millivoltsMax;
						rsp->milliampsMax = milliampsMax;
					}
					if (platGetFlags() & SWD_FLAG_CLOCK_SPEED_SETTABLE)
						rsp->maxClockRate = platGetSwdMaxClkSpeed();
					replyPayloadLen = sizeof(*rsp);
					
					if (blFlags & BL_CAP_FLAGS_NEED_PADDING)
						rsp->flags |= USB_FLAGS_NEED_PACKET_PADDING;
					break;
				}

				case SWD_COMMS_CMD_ATTACH: {
					struct SwdCommsAttachRespPacket* rpp = (struct SwdCommsAttachRespPacket*)cp->payload;
					uint8_t err;
					
					if (cmdPayloadLen)
						break;
					
					rpp->cortexType = 0;
					rpp->flags = 0;
					rpp->error = 0;
					
					wireInit();
					wireSwdSendKey();
					
					err = swdAttach(&ahbApIdx, WRAP_UNALIGNED_POINTER_32(&rpp->targetid));
					if (err != SWD_ATTACH_RET_OK)
						rpp->error = ERR_FLAG_TYPE_SWD | err;
					else {
						err = memapInit(ahbApIdx, WRAP_UNALIGNED_POINTER_32(&rpp->romTableBase));
						if (err != MEMAP_INIT_OK)
							rpp->error = ERR_FLAG_TYPE_MEMAP | err;
						else {
							err = cortexInit(WRAP_UNALIGNED_POINTER_16(&rpp->cortexType));
							if (err != CORTEX_INIT_OK) {
								rpp->cortexType = 0;
								rpp->error = ERR_FLAG_TYPE_CORTEX | err;
							}
							else {
								if (cortexHaveFpu())
									rpp->flags |= SWD_FLAG_HAS_FPU;
							}
						}
					}
					
					replyPayloadLen = sizeof(struct SwdCommsAttachRespPacket);
					break;
				}
				
				case SWD_COMMS_CMD_MEM_READ: {
					struct SwdCommsMemPacket* rpp = (struct SwdCommsMemPacket*)cp->payload;
					uint32_t numWords = UNALIGNED_16(&rpp->numWords);
					
					if (cmdPayloadLen != sizeof(struct SwdCommsMemPacket))
						break;
					
					replyPayloadLen = sizeof(struct SwdCommsMemPacket);
					if (numWords && (numWords <= maxWordsTxRx) && memapSetAddr(UNALIGNED_32(&rpp->addr)) && memapReadMultiple(WRAP_UNALIGNED_POINTER_32(rpp->words), numWords))
						replyPayloadLen += rpp->numWords * sizeof(uint32_t);
					else
						UNALIGNED_16(&rpp->numWords) = SWD_MEM_NUM_WORDS_ERROR;
					break;
				}

				case SWD_COMMS_CMD_MEM_WRITE: {
					struct SwdCommsMemPacket* rpp = (struct SwdCommsMemPacket*)cp->payload;
					uint32_t numWords = UNALIGNED_16(&rpp->numWords);
					
					if (cmdPayloadLen < sizeof(struct SwdCommsMemPacket))
						break;
					
					
					replyPayloadLen = sizeof(struct SwdCommsMemPacket);
					if (numWords == SWD_COMMS_MAX_XFER_WORDS_NO_ACK){
						replyPayloadLen = PACKET_RX_FAIL;
						numWords = maxWordsTxRx;
					}
					
					if (!numWords || numWords > maxWordsTxRx)
						numWords = SWD_MEM_NUM_WORDS_ERROR;
					else if (cmdPayloadLen != sizeof(struct SwdCommsMemPacket) + sizeof(uint32_t) * numWords)
						numWords = SWD_MEM_NUM_WORDS_ERROR;
					else if (!memapSetAddr(UNALIGNED_32(&rpp->addr)) || !memapWriteMultiple(WRAP_UNALIGNED_POINTER_32(rpp->words), numWords))
						numWords = SWD_MEM_NUM_WORDS_ERROR;

					UNALIGNED_16(&rpp->numWords) = numWords;
					break;
				}

				case SWD_COMMS_CMD_REGS_READ: {
					struct SwdCommsRegsPacket* rpp = (struct SwdCommsRegsPacket*)cp->payload;
					
					if (cmdPayloadLen != (int8_t)sizeof(struct SwdCommsRegsPacket))
						break;
					
					replyPayloadLen = sizeof(struct SwdCommsRegsPacket) + sizeof(uint32_t) * NUM_REGS;
					switch (rpp->regSet) {
						case SWD_COMMS_REG_SET_BASE:
							for (i = 0; i < NUM_REGS; i++)
								if (!cortexRegRead(i, WRAP_UNALIGNED_POINTER_32(rpp->regs + SWD_REGS_NUM_Rx(i) % NUM_REGS)))
									rpp->regSet = SWD_COMMS_REG_SET_ERROR;
							break;

						case SWD_COMMS_REG_SET_CTRL:
							if (!cortexRegRead(CORTEX_REG_XPSR, WRAP_UNALIGNED_POINTER_32(rpp->regs + SWD_REGS_NUM_XPSR % NUM_REGS)) ||
									!cortexRegRead(CORTEX_REG_MSP, WRAP_UNALIGNED_POINTER_32(rpp->regs + SWD_REGS_NUM_MSP % NUM_REGS)) ||
									!cortexRegRead(CORTEX_REG_PSP, WRAP_UNALIGNED_POINTER_32(rpp->regs + SWD_REGS_NUM_PSP % NUM_REGS)) ||
									!cortexRegRead(CORTEX_REG_CFBP, WRAP_UNALIGNED_POINTER_32(rpp->regs + SWD_REGS_NUM_CFBP % NUM_REGS)) ||
									!cortexRegRead(CORTEX_REG_FPCSR, WRAP_UNALIGNED_POINTER_32(rpp->regs + SWD_REGS_NUM_FPCSR % NUM_REGS)))
								rpp->regSet = SWD_COMMS_REG_SET_ERROR;
							break;
						
						case SWD_COMMS_REG_SET_FP1:
							j = 16;
							//fallthrough
							
						case SWD_COMMS_REG_SET_FP0:
							for (i = 0; i < NUM_REGS; i++)
								if (!cortexRegRead(CORTEX_REG_Sx(j + i), WRAP_UNALIGNED_POINTER_32(rpp->regs + SWD_REGS_NUM_Sx(i) % NUM_REGS)))
									rpp->regSet = SWD_COMMS_REG_SET_ERROR;
							break;
						default:
							rpp->regSet = SWD_COMMS_REG_SET_ERROR;
							break;
					}
					break;
				}

				case SWD_COMMS_CMD_REGS_WRITE: {
					struct SwdCommsRegsPacket* rpp = (struct SwdCommsRegsPacket*)cp->payload;
					
					if (cmdPayloadLen != (int8_t)sizeof(struct SwdCommsRegsPacket) + (int8_t)sizeof(uint32_t) * NUM_REGS)
						break;
					
					replyPayloadLen = sizeof(struct SwdCommsRegsPacket);
					switch (rpp->regSet) {
						case SWD_COMMS_REG_SET_BASE:
							for (i = 0; i < NUM_REGS; i++)
								if (!cortexRegWrite(i, UNALIGNED_32(&rpp->regs[i])))
									rpp->regSet = SWD_COMMS_REG_SET_ERROR;
							break;
						
						case SWD_COMMS_REG_SET_CTRL:
							if (!cortexRegWrite(CORTEX_REG_XPSR, UNALIGNED_32(&rpp->regs[0])) ||
									!cortexRegWrite(CORTEX_REG_MSP, UNALIGNED_32(&rpp->regs[1])) ||
									!cortexRegWrite(CORTEX_REG_PSP, UNALIGNED_32(&rpp->regs[2])) ||
									!cortexRegWrite(CORTEX_REG_CFBP, UNALIGNED_32(&rpp->regs[3])) ||
									!cortexRegWrite(CORTEX_REG_FPCSR, UNALIGNED_32(&rpp->regs[4])))
								rpp->regSet = SWD_COMMS_REG_SET_ERROR;
							break;
						
						case SWD_COMMS_REG_SET_FP1:
							j = 16;
							//fallthrough
							
						case SWD_COMMS_REG_SET_FP0:
							for (i = 0; i < NUM_REGS; i++)
								if (!cortexRegWrite(CORTEX_REG_Sx(j + i), UNALIGNED_32(&rpp->regs[i])))
									rpp->regSet = SWD_COMMS_REG_SET_ERROR;
							break;
						default:
							rpp->regSet = SWD_COMMS_REG_SET_ERROR;
					}
					break;
				}

				case SWD_COMMS_CMD_GO: {
					if (cmdPayloadLen)
						break;
						
					replyPayloadLen = 1;
					
					cp->payload[0] = cortexGo();
					break;
				}
				
				case SWD_COMMS_CMD_RESET: {
					if (cmdPayloadLen)
						break;
					
					replyPayloadLen = 1;
					
					cp->payload[0] = cortexReset();
					cortexGetStopReason();	//just to clear the bits
					break;
				}
				
				case SWD_COMMS_CMD_STOP:{
					if (cmdPayloadLen)
						break;
					
					replyPayloadLen = 1;
					
					if (cortexStop())
						goto send_stop_reason;
					
					cp->payload[0] = CORTEX_W_FAIL;
					break;
				}
				
				case SWD_COMMS_CMD_IS_STOPPED: {
					if (cmdPayloadLen)
						break;
						
					replyPayloadLen = 1;
					
			send_stop_reason:
					cp->payload[0] = cortexGetStopReason();
					break;
				}
				
				case SWD_COMMS_CMD_SINGLE_STEP: {
					if (cmdPayloadLen)
						break;
					
					replyPayloadLen = 1;
					
					if (cortexStep())
						goto send_stop_reason;
					cp->payload[0] = CORTEX_W_FAIL;
					break;
				}
				
				case SWD_COMMS_SWD_WIRE_BUS_R: {
					struct SwdCommsWireBusPacket* rpp = (struct SwdCommsWireBusPacket*)cp->payload;
					
					
					if (cmdPayloadLen != sizeof(struct SwdCommsWireBusPacket))
						break;
					
					rpp->returnVal = llWireBusRead(rpp->ap, rpp->a23, WRAP_UNALIGNED_POINTER_32(&rpp->val));
					replyPayloadLen = cmdPayloadLen;
					break;
				}
				
				case SWD_COMMS_SWD_WIRE_BUS_W: {
					struct SwdCommsWireBusPacket* rpp = (struct SwdCommsWireBusPacket*)cp->payload;
					
					
					if (cmdPayloadLen != sizeof(struct SwdCommsWireBusPacket))
						break;
					
					rpp->returnVal = llWireBusWrite(rpp->ap, rpp->a23, UNALIGNED_32(&rpp->val));
					replyPayloadLen = cmdPayloadLen;
					break;
				}
								
				case SWD_TRACE_LOG_READ: {
					uint32_t val, addr = UNALIGNED_32(cp->payload);
					uint32_t pos = 0;
					
					if (cmdPayloadLen != sizeof(uint32_t))
						break;
					
					while(pos < maxPkt - sizeof(struct CommsPacket) && memapReadAddr(addr, WRAP_UNALIGNED_POINTER_32(&val)) && (val >> 31)) {
						memapWriteAddr(addr, 0);
						cp->payload[pos++] = val;
					}
					
					replyPayloadLen = pos;
					break;
				}
	
				case SWD_POWER_CTRL: {
					
					if (cmdPayloadLen == 1) {	//simple-style power control
						cp->payload[0] = (platGetFlags() & PWR_FLAG_PWR_CTRL_ON_OFF) && platPowerOnOffSet(!!cp->payload[0]);
					}
					else if (cmdPayloadLen == 2) {	//new-style varible power control
						
						uint32_t mv = UNALIGNED_16(cp->payload);
						
						cp->payload[0] = (platGetFlags() & PWR_FLAG_PWR_CTRL_SETTABLE) && platPowerVaribleSet(mv);
					}
					else {
						
						//error - some kind of power contorl we do not support...
						break;
					}
					
					replyPayloadLen = 1;
					break;
				}
	
				case SWD_COMMS_CMD_SET_CLOCK: {
					if (platGetFlags() & SWD_FLAG_CLOCK_SPEED_SETTABLE) {
						if (cmdPayloadLen != 4)
							break;
						UNALIGNED_32(cp->payload) = platSetSwdClockSpeed(UNALIGNED_32(cp->payload))	;
						replyPayloadLen = 4;
					}
					break;
				}
				
				//if we get a "boot app", reply to help upload tool feel better
				case COMMS_CMD_BOOT_APP:
					cp->payload[0] = 1;
					replyPayloadLen = 1;
					break;

				//if we get a "get bl info packet", reboot to bootloader to allow easier update
				case COMMS_CMD_GET_INFO:
					platDeinit();
					bootloader(true);
					break;

				default:
					//ps(PSTR("PKT 0x"));
					//px(cp->cmd, 2, true);
					break;
			}
			
			//order may matter here (avr has opposite order)
			if (replyPayloadLen != PACKET_RX_FAIL)
				packetSend(cp, replyPayloadLen);
			packetRxRelease();
		}
	}


	while(1);
}

